/***************************************************************************
 *   Copyright (C) 2006 by Yoshihiro Ota
 *   ota@j.email.ne.jp
 *                                                                         
 *   This program is free software; you can redistribute it and/or modify  
 *   it under the terms of the GNU General Public License as published by  
 *   the Free Software Foundation; either version 2 of the License, or     
 *   (at your option) any later version.                                   
 *                                                                         
 *   This program is distributed in the hope that it will be useful,       
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         
 *   GNU General Public License for more details.                          
 *                                                                         
 *   You should have received a copy of the GNU General Public License     
 *   along with this program; if not, write to the                         
 *   Free Software Foundation, Inc.,                                       
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             
 *
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h> /* vfprintf */
#include <string.h>
#include <errno.h> /* errno */
#include <time.h> /* difftime */
#include <math.h> /* ceil */

#include <sys/param.h> /* MAXPATHLEN */

#include <sys/types.h>
#include <sys/stat.h>

#include <unistd.h> /* link/unlink */ /* *dir */

#include <utime.h> /* utime */
#include <dirent.h> /* dirent */

typedef enum { false, true } boolean;

#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

#define Y(x)	fprintf(stderr, "%s: " #x " %s\n", __func__, x)

// #define __DEBUG__
#if defined(__DEBUG__)
#define aprintf(txt, ...) printf("%s:" txt, __func__, __VA_ARGS__)
#define X(x)	fprintf(stderr, #x " %d\n", x)
#else
#define aprintf(txt, ...)
#define X(x)
#endif

#include <md5.h>

#ifndef __sun
#include <sha.h>
#include <sha256.h>
#endif

int busy_fs = 0;
int execute = 0;

enum {
	shell = -10, /* shell script */
	silent = 0,
	transition = 1,
} logging = silent;

typedef struct path_x {
	char *name;
	struct utimbuf time;
	struct path_x *next; /* stack: first-in, last-out */
} path_x;

typedef struct inode_x {
	int st_ino;
	char head[512];
	int head_len;
	char tail[512];
	int tail_len;
#ifndef __sun
	char md5sum[33];
	char sha1[41];
	char sha256[65];
#else
	unsigned char md5sum[16];
#endif
	struct inode_x *next;
} inode_x;

typedef struct {
	path_x *path;
	struct stat stat;
	boolean stat_checked;
	inode_x *inode;
} file_x;

typedef struct {
	dev_t st_dev; /* device ID */
	int st_size; /* file size */

	int size; /* for files */
	int max; /* for files */
	file_x **files; /* array of pointers */
	inode_x *inode; /* linkted-list stack */
} domain_x; /* device data */


typedef struct {
	int size; /* for domain */
	int max; /* for domain */
	domain_x **domain;
	path_x *access_time; /* stack of files and dirs : it maintains a list
		of directories and files to restore timestamps */
} system_x;

system_x sys = { 0, 512, NULL, NULL };

int info(int level, const char *format, ...)
{
	va_list args;
	int len;

	va_start(args, format);
	switch(level)
	{
		transition:
		default:
			len =  vfprintf(stderr, format, args);
	}
	va_end(args);
	return len;
}

domain_x* new_domain(dev_t st_dev, off_t size)
{
	domain_x *domain;
	domain = calloc(1, sizeof(*domain));
X(sizeof(*domain));
	domain->st_dev = st_dev;
	domain->st_size = size;
	domain->size = 0;
	domain->max = 16;
	domain->files = realloc(domain->files,
		domain->max * sizeof(*domain->files));
X(sizeof(*domain->files));
	aprintf("new st_dev(%d, %d)\n", st_dev, size);
	return domain;
}

domain_x* find_domain(file_x *file)
{
	int i;
	for(i = 0; i < sys.size; i++)
		if(sys.domain[i]->st_dev == file->stat.st_dev &&
			sys.domain[i]->st_size == file->stat.st_size)
			return sys.domain[i];
	if(sys.size >= sys.max)
	{
		sys.domain =
			realloc(sys.domain,
				2 * sys.max * sizeof(sys.domain));
X(sizeof(sys.domain));
		sys.max = 2 * sys.max;
	}
	sys.domain[i] = new_domain(file->stat.st_dev, file->stat.st_size);
	sys.size++;
	return sys.domain[i];
}

inode_x* find_inode(domain_x *domain, ino_t st_ino)
{
	inode_x *inode = domain->inode;
	while(inode != NULL)
		if(st_ino == inode->st_ino)
			return inode;
		else
			inode = inode->next;
	return NULL;
}

int add_new_path(path_x *access_time)
{
	access_time->next = sys.access_time;
	sys.access_time = access_time;
	return 1;
}

int restore_timestamps()
{
	path_x *path;
	while(path = sys.access_time)
	{
		if(path->name) /* restore directory timestamps */
		{
			utime(path->name, &path->time);
			free(path->name);
		}
		sys.access_time = path->next;
		free(path);
	}
	free(sys.domain);
}

int add_new_file(file_x *file)
{
	int i;
	domain_x * domain;
	domain = find_domain(file);

// Y(file->path);
	/* XXX: check if NULL */
	if(domain->size >= domain->max)
	{
		domain->files =
			realloc(domain->files,
				2 * domain->max * sizeof(domain->files));
X(sizeof(domain->files));
		domain->max = 2 * domain->max;
	}
	domain->files[domain->size++] = file;
	return 1;
}

int free_file(file_x **file)
{
	free((*file)->path->name);
	(*file)->path->name = NULL;
	free(*file);
	*file = NULL;
}

int compar(const void *data1, const void *data2)
{
	const file_x * file1 = *(file_x * const*) data1;
	const file_x * file2 = *(file_x * const*) data2;

#if defined(DEBUG)
	printf("compare 1%s -> 2%s\n",
		file1->path,
		file2->path);
#endif
#if 0
	if(file1->stat.st_size < file2->stat.st_size)
		return -1;
	else if(file1->stat.st_size > file2->stat.st_size)
		return 1;
/* compare uid and gid here so that sort does more work */
#endif
	return ceil(difftime(file1->stat.st_mtime, file2->stat.st_mtime));
}

int sort(domain_x *domain)
{
	qsort(domain->files, domain->size, sizeof(domain->files[0]), compar);
}

int file_not_linked(file_x *one, file_x *another)
{ /* if file is already processed, it will be NULL */
	if(one == NULL || another == NULL)
		return 0; /* false */
	return 1; /* true */
}

int file_stat_allowed(file_x *one, file_x *another)
{
	if(one->stat.st_dev != another->stat.st_dev)
		return 0; /* false */
	if(one->stat.st_uid != another->stat.st_uid)
		return 0; /* false */
	if(one->stat.st_gid != another->stat.st_gid)
		return 0; /* false */
	return 1; /* true */
}

/* check if the file is modified since the program is started.  If so, this
file is activly being modified and ignored from operation. */
boolean file_not_modified(file_x *file)
{
	struct stat stat;
	if(file->stat_checked)
		return true;
	if(lstat(file->path->name, &stat) == -1)
	{
		perror("file_not_modified:lstat");
		printf("path %s\n", file->path);
		return false;
	}
	if(ceil(difftime(file->stat.st_mtime, stat.st_mtime)) == 0)
	{
		file->stat_checked = true;
		return true;
	}
	return false;
}

int files_not_modified(file_x *one, file_x *another)
{
	if(!file_not_modified(one))
	{
		perror("file is modified while operation");
		return 0; /* false */
	}
	if(!file_not_modified(another))
	{
		perror("file is modified while operation");
		return 0; /* false */
	}
	return 1; /* true */
}

int read_head(domain_x *domain, file_x *file)
{
	FILE *fd;
	if(file->inode == NULL)
		file->inode = find_inode(domain, file->stat.st_ino);
	if(file->inode) /* already inode and head is read */
		return 1; /* true */
	if(!(file->inode = calloc(1, sizeof(*(file->inode)))))
	{
X(sizeof(*(file->inode)));
		perror("get_content:calloc");
		// return errno;
		return 0; /* false */
	}
	fd = fopen(file->path->name, "r");
	if(fd == NULL)
	{
		perror("get_content:fopen");
		return 0; /* false */
	}
	file->inode->head_len =
		fread(file->inode->head, 1, sizeof(file->inode->head), fd);
	fclose(fd);
	file->inode->next = domain->inode;
	domain->inode = file->inode;
	return file->inode->head_len; /* true */
}

int same_head(domain_x *domain, file_x *one, file_x *another)
{
	if(!read_head(domain, one))
		return 0;
	if(!read_head(domain, another))
		return 0;
	if(one->inode->head_len != another->inode->head_len)
		return 0;
	return !memcmp(one->inode->head, another->inode->head,
		one->inode->head_len);
}

int read_sig(file_x *file)
{
	FILE *fd;
	MD5_CTX md5;
#ifndef __sun
	SHA1_CTX sha1;
	SHA256_CTX sha256;
#endif
	int len = 0;
	if(file->inode->tail_len != 0)
		return 1; /* true */
	fd = fopen(file->path->name, "r");
	if(fd == NULL)
	{
		perror("read_sig:fopen");
		return 0; /* false */
	}
	MD5Init(&md5);
#ifndef __sun
	SHA1_Init(&sha1);
	SHA256_Init(&sha256);
#endif
	do
	{
		file->inode->tail_len = len;
		len = fread(file->inode->tail, 1, sizeof(file->inode->tail), fd);
		MD5Update(&md5, file->inode->tail, len);
#ifndef __sun
		SHA1_Update(&sha1, file->inode->tail, len);
		SHA256_Update(&sha256, file->inode->tail, len);
#endif
	}while(!feof(fd));
	if(len != 0)
		file->inode->tail_len = len;
#ifndef __sun
	MD5End(&md5, file->inode->md5sum);
	SHA1_End(&sha1, file->inode->sha1);
	SHA256_End(&sha256, file->inode->sha256);
#else
	MD5Final(file->inode->md5sum, &md5);
#endif
	fclose(fd);
#if 0
	aprintf("path(%s) -> md5(%s)\n", file->path, file->inode->md5sum);
#endif
	return 1; /* true */
}

int same_contents(file_x *one, file_x *another)
{
	read_sig(one);
	read_sig(another);
	if(one->inode->tail_len != another->inode->tail_len)
		return 0; /* false */
aprintf("same_tail_len %d:%d\n", one->inode->tail_len, another->inode->tail_len);
	if(strcmp(one->inode->tail, another->inode->tail) != 0)
		return 0; /* false */
aprintf("same_tail\n", 0);
	if(memcmp(one->inode->md5sum, another->inode->md5sum,
		sizeof(one->inode->md5sum)) != 0)
		return 0; /* false */
#ifndef __sun
	if(memcmp(one->inode->sha1, another->inode->sha1,
		sizeof(one->inode->sha1)) != 0)
		return 0; /* false */
	if(memcmp(one->inode->sha256, another->inode->sha256,
		sizeof(one->inode->sha256)) != 0)
		return 0; /* false */
#endif
aprintf("same_rig\n", 0);
	return 1; /* true */
}

int same_file(domain_x *domain, int one, int another)
{
	if(!file_not_linked(domain->files[one], domain->files[another]))
		return 0; /* false */

	if(domain->files[one]->stat.st_ino ==
		domain->files[another]->stat.st_ino) /* ignore same inode */
	{
		free_file(&domain->files[another]);
		return 0; /* it is same file but already has same inode */
	}

	if(!file_stat_allowed(domain->files[one], domain->files[another]))
		return 0; /* false */
	if(busy_fs)
		if(!files_not_modified(domain->files[one],
			domain->files[another]))
			return 0; /* false */
	if(!same_head(domain, domain->files[one], domain->files[another]))
		return 0; /* false */
aprintf("same_head\n", 0);
	if(!same_contents(domain->files[one], domain->files[another]))
		return 0; /* false */
aprintf("same_contents\n", 0);
	return 1;
}

int link_file(domain_x *domain, int one, int another)
{
	if(execute)
	{
		unlink(domain->files[another]->path->name);
		link(domain->files[one]->path->name,
			domain->files[another]->path->name);
	}

	if(logging == shell)
		printf("rm %s; ln %s, %s\n",
			domain->files[another]->path->name,
			domain->files[one]->path->name,
			domain->files[another]->path->name);
	else if(logging > 3)
		printf("unlink(%s); link(%s[%u], %s[%u])\n",
			domain->files[another]->path->name,
			domain->files[one]->path->name,
			domain->files[one]->stat.st_ino,
			domain->files[another]->path->name,
			domain->files[another]->stat.st_ino);
	free_file(&domain->files[another]);
}

void free_domain(domain_x *domain)
{
	int i;
	inode_x *inode;
	for(i = 0; i < domain->size; i++)
		if(domain->files[i] != NULL)
		{
			/* restore time stamp here */
			utime(domain->files[i]->path->name,
				&domain->files[i]->path->time);
			free(domain->files[i]->path->name);
			free(domain->files[i]);
		}
	while(domain->inode)
	{
		inode = domain->inode;
		domain->inode = domain->inode->next;
		free(inode);
	}
	free(domain);
}

int ffusion_by_domain(domain_x *domain)
{
	int i, j;
/* saved this for later enhancement.
	while(end < domain->size && domain->files[end]->stat.st_size == size)
		end++;
*/
	for(i = 0; i < domain->size; i++)
	{
		if(domain->files[i] != NULL)
		{
			for(j = i + 1; j < domain->size; j++)
			{
				if(domain->files[j] != NULL)
				{
					if(same_file(domain, i, j))
						link_file(domain, i, j);
				}
			}
		}
	}
	free(domain);
	return i;
}

int ffusion()
{
	int i = 0;
	for(i = 0; i < sys.size; i++)
	{
		file_x *file = sys.domain[i]->files[0];
		sort(sys.domain[i]);
#if 0
		aprintf("dev%u ino%u uid%u gid%u mtime%s size%u path%s\n",
			file->stat.st_dev, file->stat.st_ino,
			file->stat.st_uid, file->stat.st_gid,
			ctime(&file->stat.st_mtime), file->stat.st_size,
			file->path
			);
#endif
		ffusion_by_domain(sys.domain[i]);
	}
	restore_timestamps();
}


int loopdir(char *path,
	int (*callback)(char *parent, struct dirent *entry, void *data),
	void *data)
{
	DIR *dir;
     	struct dirent *entry;
	int i = 0;

	dir = opendir(path);
	if(dir != NULL)
	{
		while((entry = readdir(dir)) != NULL)
		{
			if(entry->d_name[0] == '.')
			{
				if(entry->d_name[1] == 0)
					continue;
				else if(entry->d_name[1] == '.' &&
					entry->d_name[2] == 0)
					continue;
			}
			callback(path, entry, data);
		}
		closedir(dir);
	}
}


int retreave(char *parent, struct dirent *entry, void *data)
{
	char path[MAXPATHLEN];
	path_x *orig_time;
	struct stat stat;

X(sizeof(*orig_time));
	orig_time = calloc(1, sizeof(*orig_time));
	if(entry)
		snprintf(path, sizeof(path), "%s/%s", parent, entry->d_name);
	else
		snprintf(path, sizeof(path), "%s", parent);

	if(lstat(path, &stat) == -1)
	{
		perror("lstat");
		return 99;
	}
	orig_time->name = strdup(path);
	orig_time->time.actime = stat.st_atime;
	orig_time->time.modtime = stat.st_mtime;
	add_new_path(orig_time);
	if(S_ISREG(stat.st_mode))
	{
		file_x *file;

		aprintf("file %s\n", orig_time->path);
		if(!(file = calloc(1, sizeof(*file))))
		{
			perror("main:calloc");
			return errno;
		}
		file->path = orig_time;
		file->stat = stat;
		add_new_file(file);
	}
	else if(S_ISDIR(stat.st_mode))
	{
		aprintf("dir %s\n", orig_time->path);
		loopdir(orig_time->name, retreave, NULL);
	}
	return 0;
}

int main(int argc, char *argv[])
{
	int i;
	char c;

	while((c = getopt(argc, argv, "bnqsvVeh")) != -1)
	{
		switch(c)
		{
			case 'b':
				busy_fs = 1;
				break;
			case 'e':
				execute = 1;
				break;
			case 'n':
			case 'q':
				logging = silent;
				break;
			case 's':
				logging = shell;
				break;
			case 'v':
				logging++;
				break;
			case 'V':
				logging--;
				break;
			case 'h':
	fprintf(stderr, "ffunsion <files>\n"
		"-b : busy file system where files are being updated\n"
		"-e : execute\n"
		"-n : silent\n"
		"-q : silent\n"
		"-s : shell script executable output\n"
		"-v : one more verbouse\n"
		"-V : one less verbouse\n"
		"-h : help\n"
		"");
				exit(0);
			case '?':
			default:
				break;
		}
	}

	argc -= optind;
	argv += optind;

	sys.domain = realloc(sys.domain, sys.max * sizeof(*sys.domain));
#if 0
	while(1)
	{
		char *newline;
		file_x *file;

		if(!(file = calloc(1, sizeof(*file))))
		{
			perror("main:calloc");
			return errno;
		}
		readline:
		if(!fgets(file->path->name, sizeof(file->path->name), stdin))
			break;

		if(newline = strchr(file->path, '\n'))
			*newline = 0; /* null-terminate */
		if(lstat(file->path, &file->stat) == -1)
		{
			perror("lstat");
			continue;
		}
		if(S_ISREG(file->stat.st_mode))
			add_new_file(file);
		else
			goto readline; /* non-file; read the next line */
		aprintf("dev%u ino%u uid%u gid%u size%u path%s mtime%s",
			file->stat.st_dev,
			file->stat.st_ino,
			file->stat.st_uid,
			file->stat.st_gid,
			file->stat.st_size,
			file->path,
			ctime(&file->stat.st_mtime)
			);
	}
#else
	for(c = 0; c < argc; c++)
	{
		retreave(argv[c], NULL, NULL);
	}
#endif
	info(transition, "gathered all filenames\n");
	ffusion();
	info(transition, "completed all files\n");
	return 0;
}
